/* 
        MmsWindow.cc

        MASL/M-MASL Copyright (c) KUNI/TDU
        Windows version Copyright (c) 1995, 1996 Tomoatsu Shimada  (shimarin)
        X Window System version Copyright (c)1997 Yamahata Isaku (ISA-9)
*/
#include <assert.h>
//#include <cassert>

#include "MmsConst.h"
#include "MmsWindow.h"

MmsWindow::MmsWindow (Display* mmsDisplay,int mmsScreen,
		      Visual* mmsVisual,GC mmsGc,int mmsDepth,
		      Window mmsWindow,
		      unsigned long mmsPalette2Pixel[NUM_OF_COLORS],
		      XMMS_OPTION* mmsXmmsOption):
  display (mmsDisplay),screen (mmsScreen),
  visual (mmsVisual),gc (mmsGc),depth (mmsDepth),
  window (mmsWindow),
  ximage (NULL),imageData (NULL),piximage (NULL),pixmapData (NULL),
  ximageOf2x0LineMode (NULL),imageDataOf2x0LineMode (NULL),
  xmmsOption (mmsXmmsOption)
{
  for (int i=0;i<NUM_OF_COLORS;i++)
    palette2Pixel[i]=mmsPalette2Pixel[i];
  
  for (int i=0;i<NUM_OF_WINDOW;i++)
    virtualWindow[i].resetVariable ();

  graphicMode=GM640x480;
  windowLeft=0;
  windowTop=0;
  windowWidth=WINDOW_SIZE_X;
  windowHeight=WINDOW_SIZE_Y;
  
  //̾ν
  int imageSizeX=WINDOW_SIZE_X;
  int imageSizeY=WINDOW_SIZE_Y;
  if (xmmsOption->isShowVirtualVram)
    {//ɽwindowvirtual vramƤɽ롣
      //Τν
      imageSizeX=VIRTUAL_VRAM_SIZE_X;
      imageSizeY=VIRTUAL_VRAM_SIZE_Y;
      XSizeHints xSizeHint;
      xSizeHint.flags=PMinSize|PMaxSize;
      xSizeHint.max_width=xSizeHint.min_width=VIRTUAL_VRAM_SIZE_X;
      xSizeHint.max_height=xSizeHint.min_height=VIRTUAL_VRAM_SIZE_Y;
      XSetWMNormalHints (display,window,&xSizeHint);
      XResizeWindow (display,window,VIRTUAL_VRAM_SIZE_X,VIRTUAL_VRAM_SIZE_Y);
      XFlush (display);
    }

  //ximageϢν
  ximage=newMaslXImage (imageSizeX,imageSizeY);
  imageData=ximage->getImagePointer ();

  //pixmapϢν
  pixmap=XCreatePixmap (display,window,imageSizeX,imageSizeY,depth);
  clearPixmap ();
  piximage=newMaslXImage (imageSizeX,imageSizeY);
  pixmapData=piximage->getImagePointer ();
  
  //ximageOfVirtualVramϢν
  ximageOfVirtualVram=NULL;
  imageDataOfVirtualVram=NULL;
  if (xmmsOption->isShowVram)
    {
      windowOfVirtualVram=
	XCreateSimpleWindow (display,RootWindow (display,screen),
			     0,0,VIRTUAL_VRAM_SIZE_X,VIRTUAL_VRAM_SIZE_Y,0,
			     BlackPixel (display,screen),
			     WhitePixel (display,screen));
      XSelectInput (display,windowOfVirtualVram,ExposureMask);
      XMapWindow (display,windowOfVirtualVram);
      XFlush (display);
      XEvent event;
      do
	{//̤mapޤԤ
	  XNextEvent (display,&event);
	}while (event.xexpose.count!=0);
      
      ximageOfVirtualVram=newMaslXImage (VIRTUAL_VRAM_SIZE_X,
					 VIRTUAL_VRAM_SIZE_Y);
      imageDataOfVirtualVram=ximageOfVirtualVram->getImagePointer ();
    }
}

MmsWindow::~MmsWindow ()
{
  delete ximage;
  XFreePixmap (display,pixmap);
  delete piximage;
  delete ximageOfVirtualVram;
  delete ximageOf2x0LineMode;
}

void MmsWindow::setGraphicMode (GraphicMode MmsGraphicMode)
{
  graphicMode=MmsGraphicMode;
  //ximagewindowΤɤΰ֤ˤɤ礭ɽ뤫ꤹ롣
  switch (graphicMode)
    {//礭
    case GM320x200:
      if (xmmsOption->isReduction)
	{
	  windowWidth=320;
	  windowHeight=200;
	}
      else
	{
	  windowWidth=320*2;
	  windowHeight=200*2;
	}
      break;
    case GM320x240:
      if (xmmsOption->isReduction)
	{
	  windowWidth=320;
	  windowHeight=240;
	}
      else
	{
	  windowWidth=320*2;
	  windowHeight=240*2;
	}
      break;
    case GM640x400: 
      windowWidth=640;
      windowHeight=400;
      break;
    case GM640x480: 
      windowWidth=640;
      windowHeight=480;
      break;
    default:
      break;//do nothing;
    }
  //window˲ɽͤ˰֤ꡣ
  windowLeft=(WINDOW_SIZE_X-windowWidth)/2;
  windowTop=(WINDOW_SIZE_Y-windowHeight)/2;

  XClearWindow (display,window);
  memset (imageData,0,WINDOW_SIZE_X*WINDOW_SIZE_Y*sizeof (char));
/*
  if (imageDataOf2x0LineMode!=NULL)
    memset (imageDataOf2x0LineMode,0,WINDOW_SIZE_X*WINDOW_SIZE_Y);
*/
  updateWindow ();
}

void MmsWindow::setWindowSize (int x1,int y1,int x2,int y2,int n=0)
{
  //////////////////////////////////window礭Ϥߤʤå
  assert (x1>=0);
  assert (x1<WINDOW_SIZE_X);
  assert (y1>=0);
  assert (y1<WINDOW_SIZE_Y);
  assert ((x1+x2)+(y1+y2)*WINDOW_SIZE_X<WINDOW_SIZE_X*WINDOW_SIZE_Y);
  if (y1+y2>=WINDOW_SIZE_Y)
    y2=WINDOW_SIZE_Y-y1-1;
  
  virtualWindow[n].vramX=x1;
  virtualWindow[n].vramY=y1;
  virtualWindow[n].width=x2;
  virtualWindow[n].height=y2;
}

void MmsWindow::moveWindow (int x,int y,int n=0)
{
  //////////////////////////////////virtual vramϤߤʤå
  assert (x>=0);
  assert (x<VIRTUAL_VRAM_SIZE_X);
  assert (y>=0);
  assert (y<VIRTUAL_VRAM_SIZE_Y);
  assert ((x+virtualWindow[n].width)+
	  (y+virtualWindow[n].height)*VIRTUAL_VRAM_SIZE_X
	  <VIRTUAL_VRAM_LENGTH);
  if (x<0)
    x=0;
  if (x>=VIRTUAL_VRAM_SIZE_X)
    x=VIRTUAL_VRAM_SIZE_X-1;
  if (y<0)
    y=0;
  if (y>=VIRTUAL_VRAM_SIZE_Y)
    y=VIRTUAL_VRAM_SIZE_Y-1;
  if (y+virtualWindow[n].height>=VIRTUAL_VRAM_LENGTH)
    virtualWindow[n].height=VIRTUAL_VRAM_SIZE_Y-y-1;
  
  virtualWindow[n].virtualVramX=x;
  virtualWindow[n].virtualVramY=y;
}

void MmsWindow::updateWindow (const unsigned char* const mmsImageData)
  //̤κmethod
{
  if (xmmsOption->isTraceMode)
    {//////trace modeäΤǤν
      for (;;)
	{
	  XEvent event;
	  XNextEvent (display,&event);
	  if (event.type!=KeyPress && event.type!=ButtonPress)
	    {
	      putImage (ximage,window);
	      continue;
	    }
	      
	  //trace modeλreturntrace modeȴ
	  if((event.type==KeyPress && 
	      XKeycodeToKeysym (display,event.xkey.keycode,0)==XK_Return)||
	     (event.type==ButtonPress && event.xbutton.button==2))
	    xmmsOption->isTraceMode=False;
	  break;
	}
    }
  putImage (mmsImageData,ximage,window);//ɽ
}

void MmsWindow::putImage (const unsigned char* const mmsImageData,
			  MaslXImage* maslXImage,Drawable drawable)
{
  char* tmpImageData=maslXImage->getImagePointer ();

  //virtual vramɽѤwindowvirtual vramƤɽ롣
  if (xmmsOption->isShowVram)
    putImageOfVirtualWindow (mmsImageData);

  if (xmmsOption->isShowVirtualVram)
    {//ɽwindowvirtual vramƤɽ롣
      if (xmmsOption->isPrivateColormapCreated)
	{//private colormapѤƤΤǡpalette==pixel value
	  //ʤΤǤΤޤͤcopyɤ
	  memcpy (tmpImageData,mmsImageData,VIRTUAL_VRAM_LENGTH);
	}
      else
	{//common colormapѤƤΤǡpaletteͤ
	  //pixel valueѴɬפͭ롣
	  for (int i=0;i<VIRTUAL_VRAM_LENGTH;i++)
	    {
	      tmpImageData [i]=palette2Pixel[mmsImageData[i]];
	    }
	}
      putImage (maslXImage,drawable);
      return;
    }
  
  /////////////dataž
  for (int i=0;i<NUM_OF_WINDOW;i++)
    {
      if (virtualWindow[i].width==0 || virtualWindow[i].height==0)
	continue;//windowĤƤ褷ʤ
      
      //ޤdataΤɤˤʤ뤫offset׻Ƥ
      int vramOffset=virtualWindow[i].vramX+
	virtualWindow[i].vramY*WINDOW_SIZE_X;
      int virtualVramOffset=virtualWindow[i].virtualVramX+
	virtualWindow[i].virtualVramY*VIRTUAL_VRAM_SIZE_X;
      assert ((virtualWindow[i].vramX+virtualWindow[i].width)+
	      (virtualWindow[i].vramY+virtualWindow[i].height)*WINDOW_SIZE_X
	      <WINDOW_SIZE_X*WINDOW_SIZE_Y);
      if (xmmsOption->isPrivateColormapCreated)
	{//private colormapѤƤΤǡpalette==pixel value
	  //ʤΤǤΤޤͤcopyɤ
	  for (int y=0;y<virtualWindow[i].height;y++)
	    {
	      memcpy (tmpImageData+vramOffset+y*WINDOW_SIZE_X,
		      mmsImageData+virtualVramOffset+y*VIRTUAL_VRAM_SIZE_X,
		      virtualWindow[i].width);
	    }
	}
      else
	{//common colormapѤƤΤǡpaletteͤ
	  //pixel valueѴɬפͭ롣
	  for (int y=0;y<virtualWindow[i].height;y++)
	    {
	      for (int x=0;x<virtualWindow[i].width;x++)
		{
		  tmpImageData[vramOffset+x+y*WINDOW_SIZE_X]=
		    palette2Pixel
		    [mmsImageData[virtualVramOffset+x+y*VIRTUAL_VRAM_SIZE_X]];
		}
	    }
	}

    }
  putImage (maslXImage,drawable);
}

void MmsWindow::putImage (MaslXImage* maslXImage,Drawable drawable)
{//////////////windowɽ
  if (xmmsOption->isShowVirtualVram)
    {//ɽwindowvirtual vramƤɽ롣
      maslXImage->XPutImage (drawable,gc,0,0,0,0,
			     VIRTUAL_VRAM_SIZE_X,VIRTUAL_VRAM_SIZE_Y);

      if (xmmsOption->isShowVirtualWindowFrame)
	{//virtual windowȤ
	  drawVirtualWindowFrame ();
	}
      return;
    }

  switch (graphicMode)
    {
    case GM320x200:
      putImage320x200 (maslXImage,drawable);
      break;
    case GM320x240:
      putImage320x240 (maslXImage,drawable);
      break;
    case GM640x400: 
      putImage640x400 (maslXImage,drawable);
      break;
    case GM640x480: 
      putImage640x480 (maslXImage,drawable);
      break;
    default:
      break;//do nothing;
    }

  if (xmmsOption->isShowVirtualWindowFrame)
    {//virtual windowȤ
      drawVirtualWindowFrame ();
    }
}

void MmsWindow::putImage320x200 (MaslXImage* maslXImage,Drawable drawable)
{
  if (xmmsOption->isReduction)
    {//򤻤window
      mmsXPutImage (maslXImage,drawable);
      return;
    }

  if (ximageOf2x0LineMode==NULL)
    allocXimageOf2x0LineMode ();//ƸƤФ줿ΤǡѴѤXImageݤ

  char* tmpImageData=maslXImage->getImagePointer ();
  //320x200line modeʤΤǺ320x200640x400˳
  for (int y=0;y<200;y++)
    {
      for (int x=0;x<320;x++)
	{
	  imageDataOf2x0LineMode[x*2+y*2*WINDOW_SIZE_X]=
	    tmpImageData[x+y*WINDOW_SIZE_X];
	  imageDataOf2x0LineMode[x*2+1+y*2*WINDOW_SIZE_X]=
	    tmpImageData[x+y*WINDOW_SIZE_X];
	}
      memcpy (imageDataOf2x0LineMode+(y*2+1)*WINDOW_SIZE_X,
	      imageDataOf2x0LineMode+(y*2)*WINDOW_SIZE_X,
	      WINDOW_SIZE_X);
    }
  ///window
  mmsXPutImage (ximageOf2x0LineMode,drawable);
}

void MmsWindow::putImage640x400 (MaslXImage* maslXImage,Drawable drawable)
{
  ////Ѵɬ̵

  ///window
  mmsXPutImage (maslXImage,drawable);
}

void MmsWindow::putImage320x240 (MaslXImage* maslXImage,Drawable drawable)
{
  if (xmmsOption->isReduction)
    {//򤻤window
      mmsXPutImage (maslXImage,drawable);
      return;
    }

  if (ximageOf2x0LineMode==NULL)
    allocXimageOf2x0LineMode ();//ƸƤФ줿ΤǡѴѤXImageݤ

  ////320x240modeʤΤǺ320x240640x480˳
  char* tmpImageData=maslXImage->getImagePointer ();
  for (int y=0;y<240;y++)
    {
      for (int x=0;x<320;x++)
	{
	  imageDataOf2x0LineMode[x*2+y*2*WINDOW_SIZE_X]=
	    tmpImageData[x+y*WINDOW_SIZE_X];
	  imageDataOf2x0LineMode[x*2+1+y*2*WINDOW_SIZE_X]=
	    tmpImageData[x+y*WINDOW_SIZE_X];
	}
      memcpy (imageDataOf2x0LineMode+(y*2+1)*WINDOW_SIZE_X,
	      imageDataOf2x0LineMode+(y*2)*WINDOW_SIZE_X,
	      WINDOW_SIZE_X);
    }

  ///window
  mmsXPutImage (ximageOf2x0LineMode,drawable);
}

void MmsWindow::putImage640x480 (MaslXImage* maslXImage,Drawable drawable)
{
  ////Ѵɬפ̵

  ///window
  mmsXPutImage (maslXImage,drawable);
}

//žѤΥ᥽å
void MmsWindow::transferDataAheadly (const unsigned char* const mmsImageData)
{
  memcpy (pixmapData,imageData,ximage->getDataLength ());
  putImage (mmsImageData,piximage,pixmap);
}

void MmsWindow::updateWindowByTransferedData ()
{
  XCopyArea (display,pixmap,window,gc,
	     windowLeft,windowTop,windowWidth,windowHeight,
	     windowLeft,windowTop);
}


void MmsWindow::allocXimageOf2x0LineMode ()
{
  ximageOf2x0LineMode=newMaslXImage (WINDOW_SIZE_X,WINDOW_SIZE_Y);
  imageDataOf2x0LineMode=ximageOf2x0LineMode->getImagePointer ();
}

void MmsWindow::putImageOfVirtualWindow
(const unsigned char* const mmsImageData)
{
  if (ximageOfVirtualVram==NULL)
    return;
  
  if (xmmsOption->isPrivateColormapCreated)
    {//private colormapѤƤΤǡpalette==pixel value
      //ʤΤǤΤޤͤcopyɤ
      memcpy (imageDataOfVirtualVram,mmsImageData,VIRTUAL_VRAM_LENGTH);
    }
  else
    {//common colormapѤƤΤǡpaletteͤ
      //pixel valueѴɬפͭ롣
      for (int i=0;i<VIRTUAL_VRAM_LENGTH;i++)
	imageDataOfVirtualVram[i]=palette2Pixel[mmsImageData[i]];
    }
  ximageOfVirtualVram->XPutImage (windowOfVirtualVram,gc,0,0,0,0,
				  VIRTUAL_VRAM_SIZE_X,VIRTUAL_VRAM_SIZE_Y);
}

void MmsWindow::drawVirtualWindowFrame ()
{
  GC lineGc=XCreateGC (display,window,0,0);
  //XSetLineAttributes (display,lineGc,1,LineSolid,CapButt,JoinMiter);
  //XSetLineAttributes (display,lineGc,1,LineOnOffDash,CapButt,JoinMiter);
  XSetLineAttributes (display,lineGc,0,LineOnOffDash,CapButt,JoinMiter);
  static XGCValues gv;
  gv.dash_offset=(gv.dash_offset+1)%4;
  XChangeGC (display,lineGc,GCDashOffset,&gv);
  XSetForeground (display,lineGc,0xffffffff);
  XSetBackground (display,lineGc,0);
  XSetFunction (display,lineGc,GXxor);
#if 0
  XSetPlaneMask(display,lineGc,
                planeMask[0]|planeMask[1]|planeMask[2]|planeMask[3]|
                planeMask[4]|planeMask[5]|planeMask[6]|planeMask[7]);
#endif

  XRectangle rectangles[NUM_OF_WINDOW];
  
  if (!xmmsOption->isShowVirtualVram)
    {
      int magnify=1;
      if ((graphicMode==GM320x200 || graphicMode==GM320x240) &&
	  !xmmsOption->isReduction)
	{//2ܤ˳礷ƤΤǡʬ礹롣
	  magnify=2;
	}

      for (int i=0;i<NUM_OF_WINDOW;i++)
	{
	  rectangles[i].x=virtualWindow[i].vramX*magnify+windowLeft;
	  rectangles[i].y=virtualWindow[i].vramY*magnify+windowTop;
	  rectangles[i].width=virtualWindow[i].width*magnify;
	  rectangles[i].height=virtualWindow[i].height*magnify;
	}
    }
  else
    {
      for (int i=0;i<NUM_OF_WINDOW;i++)
	{
	  rectangles[i].x=virtualWindow[i].virtualVramX;
	  rectangles[i].y=virtualWindow[i].virtualVramY;
	  rectangles[i].width=virtualWindow[i].width;
	  rectangles[i].height=virtualWindow[i].height;
	}
    }
  XDrawRectangles (display,window,lineGc,rectangles,NUM_OF_WINDOW);
  
  if (xmmsOption->isShowVram)
    {
      for (int i=0;i<NUM_OF_WINDOW;i++)
	{
	  rectangles[i].x=virtualWindow[i].virtualVramX;
	  rectangles[i].y=virtualWindow[i].virtualVramY;
	  rectangles[i].width=virtualWindow[i].width;
	  rectangles[i].height=virtualWindow[i].height;
	}
      XDrawRectangles (display,windowOfVirtualVram,lineGc,
		       rectangles,NUM_OF_WINDOW);
    }
  
  XFreeGC (display,lineGc);
}
